#include "Platform.h"
#include "fonts.h"
#include "mcufont.h"

#include "mf_font.c"
#include "mf_bwfont.c"
#include "mf_encoding.c"
#include "mf_justify.c"
#include "mf_kerning.c"
#include "mf_rlefont.c"
#include "mf_scaledfont.c"
#include "mf_wordwrap.c"
//mf_font.c mf_bwfont.c mf_encoding.c mf_justify.c mf_kerning.c mf_rlefont.c mf_scaledfont.c mf_wordwrap.c

#include "Colors.h"


//Simple VGA Driver 640x480x8bpp that shows embedded image
struct __using("VGA_SimpleDriver1c.spin2") vga; //RJA:  Visual Studio will show an error here, but it's OK

uint32_t ForeColor, BackColor;
char AlphaTable[256];  //one byte for every possible mix of forecolor, backcolor, and alpha

void Setup()
{
    BackColor = 2;// vga.NavyBlue;
    uint32_t* p;
    p = vga.GetPalette();
    p[BackColor] = NavyBlue;
    vga.start(VGA_BASEPIN);
    vga.CLS(BackColor);
    printf("VGA Driver started.");
    //for (;;) {}
}

typedef struct {
    const char* fontname;
    const char* filename;
    const char* text;
    bool justify;
    enum mf_align_t alignment;
    int width;
    int margin;
    int anchor;
    int scale;
} options_t;

typedef struct {
    options_t* options;
    uint8_t* buffer;
    uint16_t width;
    uint16_t height;
    uint16_t y;
    const struct mf_font_s* font;
} state_t;


static void pixel_callback(int16_t x, int16_t y, uint8_t count, uint8_t alpha, void* state)
{
    while (count--)
    {
        /* your code goes here, ex: drawPixel(x, y, alpha, color::black); */
        vga.SetPixel(x, y, alpha);// AlphaTable[alpha]);

        printf("Pixel %d,%d = %d\n", x, y, alpha);/// AlphaTable[alpha]);
        x++;
    }
}

/* Callback to render characters. */
static uint8_t character_callback(int16_t x, int16_t y, mf_char character,
    void* state)
{
    state_t* s = (state_t*)state;
    return mf_render_character(s->font, x, y, character, pixel_callback, state);
}

/* Callback to render lines. */
static bool line_callback(const char* line, uint16_t count, void* state)
{
    state_t* s = (state_t*)state;

    if (s->options->justify)
    {
        //printf("justified line\n");
        mf_render_justified(s->font, s->options->anchor, s->y,
            s->width - s->options->margin * 2,
            line, count, character_callback, state);
    }
    else
    {
        //printf("not justified line\n");
        mf_render_aligned(s->font, s->options->anchor, s->y,
            s->options->alignment, line, count,
            character_callback, state);
    }
    s->y += s->font->line_height;
    return true;
}

/* Callback to just count the lines.
 * Used to decide the image height */
bool count_lines(const char* line, uint16_t count, void* state)
{
    int* linecount = (int*)state;
    (*linecount)++;
    return true;
}



static uint8_t char_callback(int16_t x0, int16_t y0, mf_char character, void* state)
{
    return mf_render_character(&mf_rlefont_DejaVuSans12.font, x0, y0, character, &pixel_callback, state);
}

void MakeAlphaTable()
{//create table of palette entries for every possible alpha with current foreground and background colors
    float backr, backg, backb;
    float forer, foreg, foreb;
    uint32_t* p = vga.GetPalette();
    uint32_t BackColor24 = NavyBlue;// p[BackColor];
    uint32_t ForeColor24 = Yellow;// p[ForeColor];

    backr = (BackColor24 >> 16) & 0xFF;
    backg = (BackColor24 >> 8) & 0xFF;
    backb = (BackColor24 >> 0) & 0xFF;
    printf("BackColor24 = %f, %f, %f\n", backr, backg, backb);
    forer = (ForeColor24 >> 16) & 0xFF;
    foreg = (ForeColor24 >> 8) & 0xFF;
    foreb = (ForeColor24 >> 0) & 0xFF;
    printf("ForeColor24 = %f, %f, %f\n", forer, foreg, foreb);

    printf("SQRT = %d\n", vga.Root(102 * 102*256)>>4);


    int r, g, b;
    for (int a = 0; a < 256; a++)
    {
        //r = (int) ((forer * a / 256.) + (backr * (256. - a) / 256.)) &0xFF;
        //g = (int)((foreg * a / 256.) + (backg * (256. - a) / 256.)) &0xFF;
        //b = (int)((foreb * a / 256.) + (backb * (256. - a) / 256.)) &0xFF;
        r = (forer * forer * a) + (backr * backr * (256 - a));

        r = vga.Root(r) >> 4;
        b=(foreb * foreb * a) + (backb * backb * (256 - a));
        printf("b= %d\n", b);
        b = vga.Root(b) >> 4;
        printf("b= %d\n", b);
        g = (foreg * foreg * a) + (backg * backg * (256 - a));        
        g = vga.Root(g) >> 4;

        //printf("RGB = %d, %d, %d\n", r, g, b);
        //for (;;) {}

        int LowError = 256 * 256*5;
        int LowIndex = 0;
        int difr, difg, difb;

        p[a] = (r << 16) + (g << 8) + b;
        printf("p[%d] = %x\n", a, p[a]);
        
        for (int c = 0; c < 256; c++)
        {
            uint32_t pal = p[c];
            difr = (pal >> 16) & 0xFF;
            difr -= r;
            difr *= difr;
            difg = (pal >> 8) & 0xFF;
            difg -= g;
            difg *= difg;
            difb = (pal >> 0) & 0xFF;
            difb -= b;
            difb *= difb;
            int error = difr + (1*difg) + difb;
            if (error < LowError)
            {
                LowIndex = c;
                LowError = error;
            }
        }
        AlphaTable[a] = LowIndex;
        printf("Alpha[%d]= %d, RGB=%x\n", a, LowIndex, p[LowIndex]);
    }

    //for (;;) {}

}


void main()
{
    Setup();

    //ForeColor = vga.Yellow;
    MakeAlphaTable();

    vga.CLS(0);
    //for (;;) {}


    mf_render_aligned(
        &mf_rlefont_DejaVuSans12.font,
        20, 50,
        MF_ALIGN_LEFT,
        "Hello, World! This font is DejaVuSans12", 40,
        &char_callback, NULL);




    int height;
    const struct mf_font_s* font;
    struct mf_scaledfont_s scaledfont;
    options_t options;
    state_t state;// = {};



    options.fontname = "DejaVuSerif16";
    options.filename = "out.bmp";
    options.text = "This is a test of word wrapping of text";
    options.width = 100;
    options.margin = 5;
    options.scale = 1;
    options.anchor = 200;
    options.alignment = MF_ALIGN_LEFT;
    options.justify = false;


    font = mf_find_font(options.fontname);

    if (!font)
    {
        printf("No such font: %s\n", options.fontname);
        //return 2;
        for (;;) {}
    }



    if (options.scale > 1)
    {
        mf_scale_font(&scaledfont, font, options.scale, options.scale);
        font = &scaledfont.font;
    }

    /* Count the number of lines that we need. */
    height = 0;
    mf_wordwrap(font, options.width - 2 * options.margin,
        options.text, count_lines, &height);
    height *= font->height;
    height += 4;

    /* Allocate and clear the image buffer */
    state.options = &options;
    state.width = options.width;
    state.height = height;
    state.y = 200;
    state.font = font;


    /* Render the text */
    mf_wordwrap(font, options.width - 2 * options.margin,
        options.text, line_callback, &state);



    //try a second line
    //_waitms(2000); //after a couple seconds
    options.fontname = "DejaVuSerif32";
    options.filename = "out.bmp";
    options.text = "This is a test of big text";
    options.width = 600;
    options.margin = 5;
    options.scale = 1;
    options.anchor = 100;
    options.alignment = MF_ALIGN_LEFT;
    options.justify = false;

    font = mf_find_font(options.fontname);

    if (!font)
    {
        printf("No such font: %s\n", options.fontname);
        //return 2;
        for (;;) {}
    }

    /* Count the number of lines that we need. */
    height = 0;
    mf_wordwrap(font, options.width - 2 * options.margin,
        options.text, count_lines, &height);
    height *= font->height;
    height += 4;

    /* Allocate and clear the image buffer */
    state.options = &options;
    state.width = options.width;
    state.height = height;
    state.y = 100;
    state.font = font;


    /* Render the text */
    mf_wordwrap(font, options.width - 2 * options.margin,
        options.text, line_callback, &state);

    //Can we still draw text the simple way with a new font?  No.  The below uses Sans12 for some reason...
        /*
        font = mf_find_font("fixed_5x8");
        mf_render_aligned(
            font,
            20, 70,
            MF_ALIGN_LEFT,
            "Hello, World! This font is mf_bwfont_fixed_5x8", 47,
            &char_callback, NULL);
            */

            //Draw another line the long way
    options.fontname = "fixed_7x14";
    options.filename = "out.bmp";
    options.text = "This font is mf_bwfont_fixed_7x14";
    options.width = 600;
    options.margin = 5;
    options.scale = 200;  //Scale has no effect?
    options.anchor = 70;  //x
    options.alignment = MF_ALIGN_LEFT;
    options.justify = false;

    font = mf_find_font(options.fontname);

    if (!font)
    {
        printf("No such font: %s\n", options.fontname);
        //return 2;
        for (;;) {}
    }

    /* Count the number of lines that we need. */
    height = 0;
    mf_wordwrap(font, options.width - 2 * options.margin,
        options.text, count_lines, &height);
    height *= font->height;
    height += 4;

    /* Allocate and clear the image buffer */
    state.options = &options;
    state.width = options.width;
    state.height = height;
    state.y = 70;  //y
    state.font = font;


    /* Render the text */
    mf_wordwrap(font, options.width - 2 * options.margin,
        options.text, line_callback, &state);

    /*  //Eval
    #define PIN_CLK  61
    #define PIN_SS   60
    #define PIN_MOSI 59
    #define PIN_MISO 58
    */


    //Simple P2
    #define PIN_CLK  55
    #define PIN_SS   57
    #define PIN_MOSI 56
    #define PIN_MISO 54

    //Save image (308,278 bytes)
    printf("Saving BMP.\n");
    int r = 0;
    r = mount("/sd", _vfs_open_sdcardx(PIN_CLK, PIN_SS, PIN_MOSI, PIN_MISO));
    if (r != 0) {
        printf("ERROR: got error %d during mount\n", r);
    }
    chdir("/sd");
    FILE* f;
    f=fopen("alpha.bmp", "wb");
    if (!f) {
        printf("File open error.\n");
        for (;;) {}        
    }
    fwrite(vga.GetBitmap(), 1, 308278, f);
    fclose(f);
    printf("Bitmap saved.\n");


    for (;;) {}
    _waitms(100);
    vga.cls(BackColor);

        

}

